
/*******************************************************/
/* (c) Claude Michel I3S (UNSA-CNRS) 2009,2010         */
/* Distributed under GPLv3                             */
/*******************************************************/


#include <ampl_solver.h>
#include <stdio.h>
#include <stdlib.h>

abstract_solver *new_ampl_solver(char *an_ampl_solver) { return new ampl_solver(an_ampl_solver); }

int ampl_solver::init_solver(CUDFVersionedPackageList *all_versioned_packages, int other_vars) {

  nb_packages = all_versioned_packages->size();
  nb_vars = nb_packages+other_vars;

  this->all_versioned_packages = all_versioned_packages;

  nb_constraints = 0;
  usenames = false;

  coefficient = (double *)malloc(nb_vars*sizeof(double));
  index = (int *)malloc(nb_vars*sizeof(int));
  vname = (char **)malloc(nb_vars*sizeof(char *));
  solution = (double *)malloc(nb_vars*sizeof(double));

  pbfile = fopen("ampl.mod", "w");

  if ((coefficient == (double *)NULL) || 
      (index  == (int *)NULL) || 
      (vname == (char **)NULL) || 
      (solution == (double *)NULL) ||
      (pbfile == (FILE *)NULL))
    return -1;
  else
    return 0;
}

int ampl_solver::writelp(char *filename) { return 0; }

int ampl_solver::solve() {
  int status = 0;
  int rank;
  char command[1024];
  FILE *fsol = (FILE *)NULL;
  size_t size = 2048;
  char buff[2048]; char *buffer = buff; char *ptr;

  sprintf(command, "cat  ampl.mod | ( cd "AMPLDIR" ; ampl ) | tee amplsolver.out");
  system(command);

  if ((fsol = fopen("amplsolver.out", "r")) == (FILE *)NULL) {
    fprintf(stderr, "Cannot open solution file \"amplsolver.out\".\n");
    exit(-1);
  }

  for (int i = 0; i < nb_packages; i++) solution[i] = 0;

  // Get solver status
  while (! feof(fsol)) {
    getline((char **)&buffer, &size, fsol);

    if (strncmp(buffer, "solve_result_num = ", 19) == 0) {
      sscanf(buffer+19, "%d", &status);
      if ((status == 0) || (status == 101)) {
	status = 1;
	break;
      } else
	return 0;
    }
  }

  // Get objective value
  while (! feof(fsol)) {
    getline((char **)&buffer, &size, fsol);

    if (strncmp(buffer, "obj = ", 6) == 0) {
      sscanf(buffer+6, "%lf", &objval);
      break;
    }
  }

  // Get variable value
  while (! feof(fsol)) {
    getline((char **)&buffer, &size, fsol);

    if (buffer[0] == 'x') {
      sscanf(buffer+1, "%d", &rank);
      
      ptr = buffer+1;
      while (*ptr != '=') ptr++;
      ptr++;
      ptr++;
      if (*ptr == '1') solution[rank-1] = 1;
    }
  }

  return status;
}

double ampl_solver::objective_value() { return objval; }

int ampl_solver::init_solutions() { return 0; }

double ampl_solver::get_solution(CUDFVersionedPackage *package) { return solution[package->rank]; }

int ampl_solver::begin_objective(void) { 
  for (int i = 0; i < nb_vars; i++) coefficient[i] = 0;
  return 0; 
}

double ampl_solver::get_obj_coeff(CUDFVersionedPackage *package) { return coefficient[package->rank]; }

int ampl_solver::set_obj_coeff(CUDFVersionedPackage *package, double value) { coefficient[package->rank] = value; return 0; }
int ampl_solver::set_obj_coeff(int rank, double value) { coefficient[rank] = value; return 0; }

int ampl_solver::end_objective(void) {
  int starting_rank = 0;

  if (usenames) {
    for (CUDFVersionedPackageListIterator ipkg = all_versioned_packages->begin(); ipkg != all_versioned_packages->end(); ipkg++)
      fprintf(pbfile, "var %s >= 0, <= 1, binary;\n", (*ipkg)->versioned_name);
    starting_rank = nb_packages;
  }

  for (int i = starting_rank; i < nb_vars; i++)
    fprintf(pbfile, "var x%d >= 0, <= 1, binary;\n", i);

  fprintf(pbfile, "\nminimize obj: ");

  for (CUDFVersionedPackageListIterator ipkg = all_versioned_packages->begin(); ipkg != all_versioned_packages->end(); ipkg++) {
    int i = (*ipkg)->rank;
    if (coefficient[i] != 0) {
      if (usenames)
	fprintf(pbfile, " %+d*%s", (int)coefficient[i], (*ipkg)->versioned_name);
      else
	fprintf(pbfile, " %+d*x%d", (int)coefficient[i], i);
    }
  }
  for (int i = nb_packages; i < nb_vars; i++)
    if (coefficient[i] != 0) fprintf(pbfile, " %+d*x%d", (int)coefficient[i], i);
  fprintf(pbfile, ";\n\nsubject to\n\n");

  return 0;
}

int ampl_solver::begin_add_constraints(void) { 
  return 0; 
}

int ampl_solver::new_constraint(void) { rank = 0; return 0; }

double ampl_solver::get_constraint_coeff(CUDFVersionedPackage *package) {
  for (int i = 0; i < rank; i++)
    if (index[i] == package->rank) return coefficient[i];
  return 0;
}

int ampl_solver::set_constraint_coeff(CUDFVersionedPackage *package, double value) { 
  coefficient[rank] = (int)value;
  index[rank] = package->rank;
  vname[rank] = package->versioned_name;
  rank++;
  return 0;
}

int ampl_solver::set_constraint_coeff(int therank, double value) { 
  coefficient[rank] = (int)value;
  index[rank] = therank;
  vname[rank] = (char *)NULL;
  rank++;
  return 0;
}

int ampl_solver::add_constraint_geq(double bound) {
  char *sep = (char *)"";

  if (rank > 0) {
    fprintf(pbfile, "c%d: ", nb_constraints);
    for (int i = 0; i < rank; i++) {
      if (usenames && (vname[i] != (char *)NULL))
	fprintf(pbfile, "%s%+d*%s", sep, (int)coefficient[i], vname[i]);
      else
	fprintf(pbfile, "%s%+d*x%d", sep, (int)coefficient[i], index[i]);
      sep = (char *)" ";
    }
    if (bound == 0) fprintf(pbfile, " >= 0;\n"); else fprintf(pbfile, " >= %+d;\n", (int)bound);
    nb_constraints++;
  }
  return 0;
}

int ampl_solver::add_constraint_leq(double bound) {
  char *sep = (char *)"";

  if (rank > 0) {
    fprintf(pbfile, "c%d: ", nb_constraints);
    for (int i = 0; i < rank; i++) {
      if (usenames && (vname[i] != (char *)NULL))
	fprintf(pbfile, "%s%+d*%s", sep, (int)coefficient[i], vname[i]);
      else
	fprintf(pbfile, "%s%+d*x%d", sep, (int)coefficient[i], index[i]);
      sep = (char *)" ";
    }
    if (bound == 0) fprintf(pbfile, " <= 0;\n"); else fprintf(pbfile, " <= %+d;\n", (int)bound);
    nb_constraints++;
  }
  return 0;
}

int ampl_solver::add_constraint_eq(double bound) {
  char *sep = (char *)"";

  if (rank > 0) {
    fprintf(pbfile, "c%d: ", nb_constraints);
    for (int i = 0; i < rank; i++) {
      if (usenames && (vname[i] != (char *)NULL))
	fprintf(pbfile, "%s%+d*%s", sep, (int)coefficient[i], vname[i]);
      else
	fprintf(pbfile, "%s%+d*x%d", sep, (int)coefficient[i], index[i]);
      sep = (char *)" ";
    }
    if (bound == 0) fprintf(pbfile, " = 0;\n"); else fprintf(pbfile, " = %+d;\n", (int)bound);
    nb_constraints++;
  }
  return 0;
}

int ampl_solver::end_add_constraints(void) { 
  int starting_rank = 1;

  fprintf(pbfile, "\noption solver %s;\nsolve;\n", an_ampl_solver);
  fprintf(pbfile, "display solve_result_num, obj");
  if (usenames) {
    for (CUDFVersionedPackageListIterator ipkg = all_versioned_packages->begin(); ipkg != all_versioned_packages->end(); ipkg++)
      fprintf(pbfile, ", %s", (*ipkg)->versioned_name);
    starting_rank = nb_packages;
  }

  for (int i = starting_rank; i < nb_vars; i++)
    fprintf(pbfile, ", x%d", i);
  fprintf(pbfile, ";\n");

  fclose(pbfile);
  return 0; 
}
