
/*******************************************************/
/* (c) Claude Michel I3S (UNSA-CNRS) 2009,2010         */
/* Distributed under GPLv3                             */
/*******************************************************/



#include <constraint_generation.h>

//bool generate_no_lp_conflict = true;
bool generate_no_lp_conflict = false;

int new_var = 0;
vector<an_upgrade_set> upgrades;


bool is_in_provl(const CUDFVersionedPackage *pkg, CUDFProviderList *provl) {
  for (CUDFVersionedPackageListIterator ipkg = provl->begin(); ipkg != provl->end(); ipkg++)
    if ((*ipkg) == pkg) return true;
  return false;
}

bool is_in_remove(const CUDFVersionedPackage *pkg, CUDFVersionedPackageList *remove_set) {
  for (CUDFVersionedPackageListIterator ipkg = remove_set->begin(); ipkg != remove_set->end(); ipkg++)
    if ((*ipkg) == pkg) return true;
  return false;
}

bool is_in_versioned_providers(const CUDFVersionedPackage *pkg, const CUDFVersion version,
			       const CUDFVersionedProviderListIterator vpbegin, 
			       const CUDFVersionedProviderListIterator vpend) {
  CUDFVersionedProviderListIterator ivp = vpbegin;

  while (ivp != vpend) {
    if (ivp->first != version)
      for (CUDFProviderListIterator kpkg = ivp->second.begin(); kpkg != ivp->second.end(); kpkg++)
	if ((*kpkg) == pkg) return true;
    ivp++;
  }
  return false;
}


int preprocess_upgrade() {
  if (the_problem->upgrade != (CUDFVpkgList *)NULL) {
    int firstvarrank = all_packages.size();

    for (CUDFVpkgListIterator ipkgop = the_problem->upgrade->begin(); ipkgop != the_problem->upgrade->end(); ipkgop++) {
      int nb_new_var = 0;
      CUDFVersionedPackageList remove_set;
      CUDFVersionedProviderList upgrade_set;

      CUDFVirtualPackage *vpackage = (*ipkgop)->virtual_package;

      CUDFVersion highest_version = vpackage->highest_installed_provider_version;

      CUDFVersionedPackageSetIterator iverpkg = vpackage->all_versions.begin();
      CUDFVersionedPackageSetIterator iverpkgend = vpackage->all_versions.end();
      CUDFVersionedProviderListIterator iprov = vpackage->versioned_providers.begin(); 
      CUDFVersionedProviderListIterator iprovend = vpackage->versioned_providers.end(); 


      if (vpackage->highest_installed != (CUDFVersionedPackage *)NULL)
	if (vpackage->highest_installed->version > highest_version)
	  highest_version = vpackage->highest_installed->version;

      // even if not installed, it seems that we have to upgrade ...
      // If nothing is installed ... it cannot be upgraded ...
      //      if (highest_version == 0) continue;

      // Remove all the "all versions" providers
      for (CUDFProviderListIterator jprov = vpackage->providers.begin(); jprov != vpackage->providers.end(); jprov++)
	remove_set.push_back(*jprov);

      // Remove all packages with version lower than hver
      for (; (iverpkg != iverpkgend) && ((*iverpkg)->version < highest_version); iverpkg++)
	remove_set.push_back(*iverpkg);
      // The same for versioned providers
      for (; (iprov != iprovend) && (iprov->first < highest_version); iprov++)
	for (CUDFProviderListIterator kpkg = iprov->second.begin(); kpkg != iprov->second.end(); kpkg++)
	  remove_set.push_back(*kpkg);

      // Upgrade highest_version version packages ... this does not seems to be right ...
      // the highest installed version package has also to go through the filters
      /*
      {
	CUDFProviderList hverpkgs;
	if ((iverpkg != vpackage->all_versions.end()) && 
	    ((*iverpkg)->version == highest_version) &&
	    (! is_in_remove(*iverpkg, &remove_set)) &&
	    (! is_in_versioned_providers(*iverpkg, (*iverpkg)->version, iprov, iprovend))) {
	  hverpkgs.push_back(*iverpkg);
	  iverpkg++;
	}
	if ((iprov != iprovend) &&
	    (iprov->first == highest_version)) {
	  for (CUDFProviderListIterator kpkg = iprov->second.begin(); kpkg != iprov->second.end(); kpkg++)
	    if (! is_in_versioned_providers(*kpkg, iprov->first, iprov, iprovend))
	      if (! is_in_provl(*kpkg, &hverpkgs)) hverpkgs.push_back(*kpkg);
	  iprov++;
	}
	if (hverpkgs.size() > 0) {
	  upgrade_set.insert(CUDFVersionedProviderList::value_type(highest_version, hverpkgs));
	  if (hverpkgs.size() > 1) nb_new_var++;
	}
      }
      */

      a_compptr comp = get_comparator((*ipkgop)->op);

      while ((iverpkg != iverpkgend) || (iprov != iprovend)) {
	CUDFProviderList provl;
	CUDFVersion version = 0;

	if (iverpkg != iverpkgend)
	  version = (*iverpkg)->version;
	if ((iprov != iprovend) && (version < iprov->first))
	  version = iprov->first;

	if ((iverpkg != iverpkgend) && ((*iverpkg)->version == version)) {
	  if (comp(version, (*ipkgop)->version)) {
	    if (! is_in_remove(*iverpkg, &remove_set)) {
	      if (is_in_versioned_providers(*iverpkg, (*iverpkg)->version, iprov, iprovend))
		remove_set.push_back(*iverpkg);
	      else
		provl.push_back(*iverpkg);
	    }
	  } else if (! is_in_remove(*iverpkg, &remove_set))
	      remove_set.push_back(*iverpkg);
	  iverpkg++;
	}

	if ((iprov != iprovend) && (iprov->first == version)) {
	  if (comp(version, (*ipkgop)->version)) {
	    for (CUDFProviderListIterator kpkg = iprov->second.begin(); kpkg != iprov->second.end(); kpkg++)
	      if (! is_in_remove(*kpkg, &remove_set)) {
		if (is_in_versioned_providers(*kpkg, iprov->first, iprov, iprovend) || ((*kpkg)->virtual_package == vpackage))
		  remove_set.push_back(*kpkg);
		else
		  if (! is_in_provl(*kpkg, &provl)) provl.push_back(*kpkg);
	      }
	  } else {
	    for (CUDFProviderListIterator kpkg = iprov->second.begin(); kpkg != iprov->second.end(); kpkg++)
	      if (! is_in_remove(*kpkg, &remove_set))
		remove_set.push_back(*kpkg);
	  }
	  iprov++;
	}

	if (provl.size() > 0) {
	  upgrade_set.insert(CUDFVersionedProviderList::value_type(version, provl));
	  if (provl.size() > 1) nb_new_var++;
	}
      }

      //      if ((remove_set.size() != 0) || (upgrade_set.size() != 0)) {
      if (upgrade_set.size() != 0) {
	upgrades.push_back(an_upgrade_set(nb_new_var, firstvarrank, remove_set, upgrade_set));
	new_var += nb_new_var;
	firstvarrank += nb_new_var;
      } else { // We need somethin to upgrade to ...
	printf("Cannot upgrade to %s.\n", vpackage->name);
	return -1;
      }
    }
  }
  return 0;
}

int generate_constraints(abstract_solver &solver, int criteria) { 
  int nb_packages = all_packages.size();
  int nb_vpackages = all_virtual_packages.size();
  vector<bool> keep_package_handled(nb_vpackages, false);
  CUDFVirtualPackageList installed_vpkgs_mv;
  CUDFVirtualPackageList installed_vpkgs_uv;

  if (nb_packages == 0) {
    fprintf(stderr, "generate_constraints: no declared package !\n");
    exit(-1);
  }

  if (the_problem == (CUDFproblem *)NULL) {
    fprintf(stderr, "no request declared.\n");
    exit(-1);
  }

  if ((criteria == CRIT_LEXPARANOID) || 
      (criteria == CRIT_LEXTRENDY) || 
      (criteria == CRIT_LEXTRENDY2) || 
      (criteria == CRIT_PARANOID) || 
      (criteria == CRIT_TRENDY) || 
      (criteria == CRIT_TRENDY2)) {
    for (CUDFVirtualPackageListIterator vpkg = all_virtual_packages.begin(); vpkg != all_virtual_packages.end(); vpkg++) {
      if ((*vpkg)->highest_installed != (CUDFVersionedPackage *)NULL) {
	if ((*vpkg)->all_versions.size() == 1)
	  installed_vpkgs_uv.push_back((*vpkg));
	else
	  installed_vpkgs_mv.push_back((*vpkg));
      }
    }
  }

  if (preprocess_upgrade() != 0) return -1; // Trying to upgrade non existent package
  solver.init_solver(&all_packages, new_var + installed_vpkgs_mv.size());

  //----------------------------------------------------------------------------------------------------
  // Objective function

  solver.begin_objective();

  if (criteria == CRIT_CONSERVATIVE) {
    // Try to keep installed already installed packages
    if (installed_packages.size() > 0)
      for (CUDFVersionedPackageListIterator ipkg = installed_packages.begin(); ipkg != installed_packages.end(); ipkg++) 
	solver.set_obj_coeff(*ipkg, -1);
    
    // Try to keep uninstalled already packages uninstalled
    if (uninstalled_packages.size() > 0)
      for (CUDFVersionedPackageListIterator ipkg = uninstalled_packages.begin(); ipkg != uninstalled_packages.end(); ipkg++)
	solver.set_obj_coeff(*ipkg, +1);
    
    // Try upgrade if and only if required
    // WARNING : highest installed version vs providers ...
    if (upgrades.size() > 0) {
      for (vector<an_upgrade_set>::iterator iup = upgrades.begin(); iup != upgrades.end(); iup++) {
	int weight = (*iup).upgrade_set.size() + 1;
	
	for (CUDFVersionedProviderListIterator ivpl = (*iup).upgrade_set.begin(); ivpl != (*iup).upgrade_set.end(); ivpl++) {
	  for (CUDFProviderListIterator kpkg = ivpl->second.begin(); kpkg != ivpl->second.end(); kpkg++)
	    solver.set_obj_coeff(*kpkg, -weight);
	  weight--;
	}
      }
    }
  } else if (criteria == CRIT_PARANOID) { // less changes as possible ...
    // Try to keep uninstalled already packages uninstalled
    if (uninstalled_packages.size() > 0) {
      for (CUDFVersionedPackageListIterator ipkg = uninstalled_packages.begin(); ipkg != uninstalled_packages.end(); ipkg++)
	solver.set_obj_coeff(*ipkg, +1);
    }
    // Try to keep installed already installed packages ... and that's it !
    if (installed_packages.size() > 0) {
      for (CUDFVersionedPackageListIterator ipkg = installed_packages.begin(); ipkg != installed_packages.end(); ipkg++) 
	solver.set_obj_coeff(*ipkg, -(((*ipkg)->virtual_package->all_versions.size() == 1)?nb_packages:1));
    }
    // min removed for multiple versions
    for (int i = 0; i < (int)installed_vpkgs_mv.size(); i++)
      solver.set_obj_coeff(nb_packages + new_var + i, -nb_packages);
  } else if (criteria == CRIT_TRENDY2) {
    // Try to keep uninstalled already packages uninstalled
    // combines max uptodate & min new
    if (uninstalled_packages.size() > 0)
      for (CUDFVersionedPackageListIterator ipkg = uninstalled_packages.begin(); ipkg != uninstalled_packages.end(); ipkg++) {
	if ((*ipkg)->virtual_package->highest_installed != (CUDFVersionedPackage *)NULL) {
	  if ((*ipkg)->virtual_package->highest_version == (*ipkg)->version)
	    solver.set_obj_coeff(*ipkg, -1);
	  else
	    solver.set_obj_coeff(*ipkg, 1);
	} else {
	  if ((*ipkg)->virtual_package->highest_version == (*ipkg)->version)
	    solver.set_obj_coeff(*ipkg, -1);
	  else
	    solver.set_obj_coeff(*ipkg, 10);
	}
      }

    // min remove
    for (int i = 0; i < (int)installed_vpkgs_mv.size(); i++)
      solver.set_obj_coeff(nb_packages + new_var + i, -100);
    for (CUDFVirtualPackageListIterator vpkg = installed_vpkgs_uv.begin(); vpkg != installed_vpkgs_uv.end(); vpkg++)
      solver.set_obj_coeff(*((*vpkg)->all_versions.begin()), -100);

  } else if (criteria == CRIT_TRENDY) {
    // min new: Minimize the number of new packages
    if (uninstalled_packages.size() > 0) {
      for (CUDFVersionedPackageListIterator ipkg = uninstalled_packages.begin(); ipkg != uninstalled_packages.end(); ipkg++) {
	solver.set_obj_coeff(*ipkg, +1);
      }
    }

    // min notuptodate: Minimize the number of installed packages not at their latest version
    for (CUDFVersionedPackageListIterator ipkg = all_packages.begin(); ipkg != all_packages.end(); ipkg++)
      if ((*ipkg)->virtual_package->highest_version != (*ipkg)->version) {
	solver.set_obj_coeff(*ipkg, 5000 + solver.get_obj_coeff(*ipkg));
      }

    // min removed
    for (int i = 0; i < (int)installed_vpkgs_mv.size(); i++) {
      solver.set_obj_coeff(nb_packages + new_var + i, -500000);
    }
    for (CUDFVirtualPackageListIterator vpkg = installed_vpkgs_uv.begin(); vpkg != installed_vpkgs_uv.end(); vpkg++) {
      solver.set_obj_coeff(*((*vpkg)->all_versions.begin()), -500000 + solver.get_obj_coeff(*((*vpkg)->all_versions.begin())));
    }

  } else if ((criteria == CRIT_LEXPARANOID) || (criteria == CRIT_LEXTRENDY) || (criteria == CRIT_LEXTRENDY2)) {
    // min removed
    for (int i = 0; i < (int)installed_vpkgs_mv.size(); i++)
      solver.set_obj_coeff(nb_packages + new_var + i, -1);
    for (CUDFVirtualPackageListIterator vpkg = installed_vpkgs_uv.begin(); vpkg != installed_vpkgs_uv.end(); vpkg++)
      solver.set_obj_coeff(*((*vpkg)->all_versions.begin()), -1);
  } else if (criteria == CRIT_FRESHER) {
    // Try to keep installed already installed packages
    if (installed_packages.size() > 0)
      for (CUDFVersionedPackageListIterator ipkg = installed_packages.begin(); ipkg != installed_packages.end(); ipkg++) 
	solver.set_obj_coeff(*ipkg, -1);
    
    // Try to keep uninstalled already packages uninstalled
    if (uninstalled_packages.size() > 0)
      for (CUDFVersionedPackageListIterator ipkg = uninstalled_packages.begin(); ipkg != uninstalled_packages.end(); ipkg++)
	solver.set_obj_coeff(*ipkg, (*ipkg)->virtual_package->highest_version - (*ipkg)->version + 1);

    // Release installed package from install request so that a fresher version can be chosen    
    if (the_problem->install != (CUDFVpkgList *)NULL) {
      for (CUDFVpkgListIterator ipkgop = the_problem->install->begin(); ipkgop != the_problem->install->end(); ipkgop++) {
	CUDFVirtualPackage *vpackage = (*ipkgop)->virtual_package;
	a_compptr comp = get_comparator((*ipkgop)->op);
	if (vpackage->all_versions.size() > 0) // Install P = install one version of P
	  for (CUDFVersionedPackageSetIterator ipkg = vpackage->all_versions.begin(); ipkg != vpackage->all_versions.end(); ipkg++)
	    if (comp((*ipkg)->version, (*ipkgop)->version)) 
	      if ((*ipkg)->installed)
		solver.set_obj_coeff(*ipkg, (*ipkg)->virtual_package->highest_version - (*ipkg)->version + 1);
	if (vpackage->providers.size() > 0) // or install one of the providers of P
	  for (CUDFProviderListIterator jpkg = vpackage->providers.begin(); jpkg != vpackage->providers.end(); jpkg++)
	    if (solver.get_constraint_coeff(*jpkg) == 0) 
	      if ((*jpkg)->installed)
		solver.set_obj_coeff(*jpkg, (*jpkg)->virtual_package->highest_version - (*jpkg)->version + 1);
	// or install one of the providers with the right version
	for (CUDFVersionedProviderListIterator jpkg = vpackage->versioned_providers.begin(); 
	     jpkg != vpackage->versioned_providers.end(); jpkg++)
	  if (comp(jpkg->first, (*ipkgop)->version))
	    for (CUDFProviderListIterator kpkg = jpkg->second.begin(); kpkg != jpkg->second.end(); kpkg++)
	      if (solver.get_constraint_coeff(*kpkg) == 0) 
		if ((*kpkg)->installed)
		  solver.set_obj_coeff(*kpkg, (*kpkg)->virtual_package->highest_version - (*kpkg)->version + 1);
      }
    }

    // Try to upgrade to the most recent version
    // WARNING : highest installed version vs providers ...
    if (upgrades.size() > 0) {
      for (vector<an_upgrade_set>::iterator iup = upgrades.begin(); iup != upgrades.end(); iup++) {
	int weight = 1;
	int weight_inc = nb_packages;
	
	for (CUDFVersionedProviderListIterator ivpl = (*iup).upgrade_set.begin(); ivpl != (*iup).upgrade_set.end(); ivpl++) {
	  for (CUDFProviderListIterator kpkg = ivpl->second.begin(); kpkg != ivpl->second.end(); kpkg++)
	    solver.set_obj_coeff(*kpkg, -weight);
	  weight += weight_inc;
	}
      }
    }
  } else if (criteria == CRIT_SFRESHER) {
    // Try to keep installed already installed packages
    if (installed_packages.size() > 0)
      for (CUDFVersionedPackageListIterator ipkg = installed_packages.begin(); ipkg != installed_packages.end(); ipkg++) 
	solver.set_obj_coeff(*ipkg, -1);
    
    // Try to keep uninstalled already packages uninstalled
    if (uninstalled_packages.size() > 0)
      for (CUDFVersionedPackageListIterator ipkg = uninstalled_packages.begin(); ipkg != uninstalled_packages.end(); ipkg++)
	solver.set_obj_coeff(*ipkg, (*ipkg)->virtual_package->highest_version - (*ipkg)->version + 1);

    // Release installed package from install request so that a fresher version can be chosen    
    if (the_problem->install != (CUDFVpkgList *)NULL) {
      for (CUDFVpkgListIterator ipkgop = the_problem->install->begin(); ipkgop != the_problem->install->end(); ipkgop++) {
	CUDFVirtualPackage *vpackage = (*ipkgop)->virtual_package;
	a_compptr comp = get_comparator((*ipkgop)->op);
	if (vpackage->all_versions.size() > 0) // Install P = install one version of P
	  for (CUDFVersionedPackageSetIterator ipkg = vpackage->all_versions.begin(); ipkg != vpackage->all_versions.end(); ipkg++)
	    if (comp((*ipkg)->version, (*ipkgop)->version)) 
	      if ((*ipkg)->installed)
		solver.set_obj_coeff(*ipkg, (*ipkg)->virtual_package->highest_version - (*ipkg)->version + 1);
	if (vpackage->providers.size() > 0) // or install one of the providers of P
	  for (CUDFProviderListIterator jpkg = vpackage->providers.begin(); jpkg != vpackage->providers.end(); jpkg++)
	    if (solver.get_constraint_coeff(*jpkg) == 0) 
	      if ((*jpkg)->installed)
		solver.set_obj_coeff(*jpkg, (*jpkg)->virtual_package->highest_version - (*jpkg)->version + 1);
	// or install one of the providers with the right version
	for (CUDFVersionedProviderListIterator jpkg = vpackage->versioned_providers.begin(); 
	     jpkg != vpackage->versioned_providers.end(); jpkg++)
	  if (comp(jpkg->first, (*ipkgop)->version))
	    for (CUDFProviderListIterator kpkg = jpkg->second.begin(); kpkg != jpkg->second.end(); kpkg++)
	      if (solver.get_constraint_coeff(*kpkg) == 0) 
		if ((*kpkg)->installed)
		  solver.set_obj_coeff(*kpkg, (*kpkg)->virtual_package->highest_version - (*kpkg)->version + 1);
      }
    }

    // Try to upgrade to the most recent version
    // WARNING : highest installed version vs providers ...
    if (upgrades.size() > 0) {
      for (vector<an_upgrade_set>::iterator iup = upgrades.begin(); iup != upgrades.end(); iup++) {
	int weight = 1;
	int weight_inc = 1;
	
	for (CUDFVersionedProviderListIterator ivpl = (*iup).upgrade_set.begin(); ivpl != (*iup).upgrade_set.end(); ivpl++) {
	  for (CUDFProviderListIterator kpkg = ivpl->second.begin(); kpkg != ivpl->second.end(); kpkg++)
	    solver.set_obj_coeff(*kpkg, -weight);
	  weight += weight_inc;
	}
      }
    }
  } else if (criteria == CRIT_FFRESHER) {
    // Try to keep installed already installed packages
    if (installed_packages.size() > 0)
      for (CUDFVersionedPackageListIterator ipkg = installed_packages.begin(); ipkg != installed_packages.end(); ipkg++) 
	solver.set_obj_coeff(*ipkg, -1);
    
    // Try to keep uninstalled already packages uninstalled
    if (uninstalled_packages.size() > 0)
      for (CUDFVersionedPackageListIterator ipkg = uninstalled_packages.begin(); ipkg != uninstalled_packages.end(); ipkg++)
	solver.set_obj_coeff(*ipkg, ((*ipkg)->virtual_package->highest_version - (*ipkg)->version)*nb_packages + 1);

    // Release installed package from install request so that a fresher version can be chosen    
    if (the_problem->install != (CUDFVpkgList *)NULL) {
      for (CUDFVpkgListIterator ipkgop = the_problem->install->begin(); ipkgop != the_problem->install->end(); ipkgop++) {
	CUDFVirtualPackage *vpackage = (*ipkgop)->virtual_package;
	a_compptr comp = get_comparator((*ipkgop)->op);
	if (vpackage->all_versions.size() > 0) // Install P = install one version of P
	  for (CUDFVersionedPackageSetIterator ipkg = vpackage->all_versions.begin(); ipkg != vpackage->all_versions.end(); ipkg++)
	    if (comp((*ipkg)->version, (*ipkgop)->version)) 
	      if ((*ipkg)->installed)
		solver.set_obj_coeff(*ipkg, ((*ipkg)->virtual_package->highest_version - (*ipkg)->version)*nb_packages + 1);
	if (vpackage->providers.size() > 0) // or install one of the providers of P
	  for (CUDFProviderListIterator jpkg = vpackage->providers.begin(); jpkg != vpackage->providers.end(); jpkg++)
	    if (solver.get_constraint_coeff(*jpkg) == 0) 
	      if ((*jpkg)->installed)
		solver.set_obj_coeff(*jpkg, ((*jpkg)->virtual_package->highest_version - (*jpkg)->version)*nb_packages + 1);
	// or install one of the providers with the right version
	for (CUDFVersionedProviderListIterator jpkg = vpackage->versioned_providers.begin(); 
	     jpkg != vpackage->versioned_providers.end(); jpkg++)
	  if (comp(jpkg->first, (*ipkgop)->version))
	    for (CUDFProviderListIterator kpkg = jpkg->second.begin(); kpkg != jpkg->second.end(); kpkg++)
	      if (solver.get_constraint_coeff(*kpkg) == 0) 
		if ((*kpkg)->installed)
		  solver.set_obj_coeff(*kpkg, ((*kpkg)->virtual_package->highest_version - (*kpkg)->version)*nb_packages + 1);
      }
    }

    // Try to upgrade to the most recent version
    // WARNING : highest installed version vs providers ...
    if (upgrades.size() > 0) {
      for (vector<an_upgrade_set>::iterator iup = upgrades.begin(); iup != upgrades.end(); iup++) {
	int weight = 1;
	int weight_inc = nb_packages;
	
	for (CUDFVersionedProviderListIterator ivpl = (*iup).upgrade_set.begin(); ivpl != (*iup).upgrade_set.end(); ivpl++) {
	  for (CUDFProviderListIterator kpkg = ivpl->second.begin(); kpkg != ivpl->second.end(); kpkg++)
	    solver.set_obj_coeff(*kpkg, -weight);
	  weight += weight_inc;
	}
      }
    }
  } else {
    // Try to keep installed already installed packages
    if (installed_packages.size() > 0)
      for (CUDFVersionedPackageListIterator ipkg = installed_packages.begin(); ipkg != installed_packages.end(); ipkg++) 
	solver.set_obj_coeff(*ipkg, -1);
    
    // Try to keep uninstalled already packages uninstalled
    if (uninstalled_packages.size() > 0)
      for (CUDFVersionedPackageListIterator ipkg = uninstalled_packages.begin(); ipkg != uninstalled_packages.end(); ipkg++)
	solver.set_obj_coeff(*ipkg, +1);
    
    // Try to upgrade to the most recent version
    // WARNING : highest installed version vs providers ...
    if (upgrades.size() > 0) {
      for (vector<an_upgrade_set>::iterator iup = upgrades.begin(); iup != upgrades.end(); iup++) {
	int weight = 1;
	int weight_inc = nb_packages;
	
	for (CUDFVersionedProviderListIterator ivpl = (*iup).upgrade_set.begin(); ivpl != (*iup).upgrade_set.end(); ivpl++) {
	  for (CUDFProviderListIterator kpkg = ivpl->second.begin(); kpkg != ivpl->second.end(); kpkg++)
	    solver.set_obj_coeff(*kpkg, -weight);
	  weight += weight_inc;
	}
      }
    }
  }

  solver.end_objective();
  
  if (criteria == CRIT_LEXPARANOID) {
    // min change
    solver.new_objective();
    if (installed_packages.size() > 0) {
      for (CUDFVersionedPackageListIterator ipkg = installed_packages.begin(); ipkg != installed_packages.end(); ipkg++) 
	solver.set_obj_coeff(*ipkg, -1);
    }
    // Try to keep uninstalled already packages uninstalled
    if (uninstalled_packages.size() > 0) {
      for (CUDFVersionedPackageListIterator ipkg = uninstalled_packages.begin(); ipkg != uninstalled_packages.end(); ipkg++)
	solver.set_obj_coeff(*ipkg, +1);
    }
    solver.add_objective();
  }

  if (criteria == CRIT_LEXTRENDY2) {
    // max uptodate: Maximize the number of installed packages with their latest available version 
    solver.new_objective();
    for (CUDFVersionedPackageListIterator ipkg = all_packages.begin(); ipkg != all_packages.end(); ipkg++)
      if ((*ipkg)->virtual_package->highest_version == (*ipkg)->version)
	solver.set_obj_coeff(*ipkg, -1);
      else
	solver.set_obj_coeff(*ipkg, +1);
    solver.add_objective();

    // min new: Minimize the number of new packages
    solver.new_objective();
    if (uninstalled_packages.size() > 0) {
      for (CUDFVersionedPackageListIterator ipkg = uninstalled_packages.begin(); ipkg != uninstalled_packages.end(); ipkg++)
	solver.set_obj_coeff(*ipkg, +1);
    }
    solver.add_objective();
  }

  if (criteria == CRIT_LEXTRENDY) {
    // min notuptodate: Minimize the number of installed packages not at their latest version
    solver.new_objective();
    for (CUDFVersionedPackageListIterator ipkg = all_packages.begin(); ipkg != all_packages.end(); ipkg++)
      if ((*ipkg)->virtual_package->highest_version != (*ipkg)->version)
	solver.set_obj_coeff(*ipkg, +1);
    solver.add_objective();

    // min new: Minimize the number of new packages
    solver.new_objective();
    if (uninstalled_packages.size() > 0) {
      for (CUDFVersionedPackageListIterator ipkg = uninstalled_packages.begin(); ipkg != uninstalled_packages.end(); ipkg++)
	solver.set_obj_coeff(*ipkg, +1);
    }
    solver.add_objective();
  }

  //----------------------------------------------------------------------------------------------------
  // Constraints

  solver.begin_add_constraints();

  if ((criteria == CRIT_LEXPARANOID) || 
      (criteria == CRIT_PARANOID) || 
      (criteria == CRIT_TRENDY) || 
      (criteria == CRIT_TRENDY2) || 
      (criteria == CRIT_LEXTRENDY) || 
      (criteria == CRIT_LEXTRENDY2)) {
    int rank = nb_packages + new_var;
    for (CUDFVirtualPackageListIterator vpkg = installed_vpkgs_mv.begin(); vpkg != installed_vpkgs_mv.end(); vpkg++) {
      solver.new_constraint();
      for (CUDFVersionedPackageSetIterator ipkg = (*vpkg)->all_versions.begin(); ipkg != (*vpkg)->all_versions.end(); ipkg++)
	solver.set_constraint_coeff(*ipkg, +1);
      solver.set_constraint_coeff(rank, -1);
      solver.add_constraint_geq(0);
      for (CUDFVersionedPackageSetIterator ipkg = (*vpkg)->all_versions.begin(); ipkg != (*vpkg)->all_versions.end(); ipkg++) {
	solver.new_constraint();
	solver.set_constraint_coeff(*ipkg, -1);
	solver.set_constraint_coeff(rank, +1);
	solver.add_constraint_geq(0);
      }
      rank++;
    }
  }

  // Install 
  if (the_problem->install != (CUDFVpkgList *)NULL) {
    for (CUDFVpkgListIterator ipkgop = the_problem->install->begin(); ipkgop != the_problem->install->end(); ipkgop++) {
      CUDFVirtualPackage *vpackage = (*ipkgop)->virtual_package;
      a_compptr comp = get_comparator((*ipkgop)->op);
      bool has_pkg = false;

      solver.new_constraint();
      if (vpackage->all_versions.size() > 0) // Install P = install one version of P
	for (CUDFVersionedPackageSetIterator ipkg = vpackage->all_versions.begin(); ipkg != vpackage->all_versions.end(); ipkg++)
	  if (comp((*ipkg)->version, (*ipkgop)->version)) {
	    has_pkg = true;
	    solver.set_constraint_coeff(*ipkg, +1);
	  }
      if (vpackage->providers.size() > 0) // or install one of the providers of P
	for (CUDFProviderListIterator jpkg = vpackage->providers.begin(); jpkg != vpackage->providers.end(); jpkg++)
	  if (solver.get_constraint_coeff(*jpkg) == 0) {
	    has_pkg = true;
	    solver.set_constraint_coeff(*jpkg, +1);
	  }
      // or install one of the providers with the right version
      for (CUDFVersionedProviderListIterator jpkg = vpackage->versioned_providers.begin(); 
	   jpkg != vpackage->versioned_providers.end(); jpkg++)
	if (comp(jpkg->first, (*ipkgop)->version))
	  for (CUDFProviderListIterator kpkg = jpkg->second.begin(); kpkg != jpkg->second.end(); kpkg++)
	    if (solver.get_constraint_coeff(*kpkg) == 0) {
	      has_pkg = true;
	      solver.set_constraint_coeff(*kpkg, +1);
	    }
      if (has_pkg)
	solver.add_constraint_geq(+1);
      else
	return -1; // Cannot install this pkg
    }
  }

  // Remove
  if (the_problem->remove != (CUDFVpkgList *)NULL) {
    for (CUDFVpkgListIterator ipkgop = the_problem->remove->begin(); ipkgop != the_problem->remove->end(); ipkgop++) {
      CUDFVirtualPackage *vpackage = (*ipkgop)->virtual_package;
      a_compptr comp = get_comparator((*ipkgop)->op);
      solver.new_constraint();
      if (vpackage->all_versions.size() > 0) // Remove all the versions of the package
	for (CUDFVersionedPackageSetIterator ipkg = vpackage->all_versions.begin(); ipkg != vpackage->all_versions.end(); ipkg++)
	  if (comp((*ipkg)->version, ((*ipkgop)->version))) solver.set_constraint_coeff(*ipkg, +1);
      if (vpackage->providers.size() > 0) // as well as all the providers
	for (CUDFProviderListIterator jpkg = vpackage->providers.begin(); jpkg != vpackage->providers.end(); jpkg++)
	  if (solver.get_constraint_coeff(*jpkg) == 0) solver.set_constraint_coeff(*jpkg, +1);
      // as well as all the versioned providers with the right version
      for (CUDFVersionedProviderListIterator jpkg = vpackage->versioned_providers.begin(); 
	   jpkg != vpackage->versioned_providers.end(); jpkg++)
	if (comp(jpkg->first, (*ipkgop)->version))
	  for (CUDFProviderListIterator kpkg = jpkg->second.begin(); kpkg != jpkg->second.end(); kpkg++)
	    if (solver.get_constraint_coeff(*kpkg) == 0) solver.set_constraint_coeff(*kpkg, +1);
      solver.add_constraint_eq(0);
    }
  }


  // Upgrade : There can be only one VERSION of (already_installed with higher version, + accepted by criteria and > higher version) installed, 
  // all other removed
  // WARNING: Providers not handled here ...
  if (upgrades.size() > 0) {
    for (vector<an_upgrade_set>::iterator iup = upgrades.begin(); iup != upgrades.end(); iup++) {

      // Remove packages belonging to upgrades remove set
      solver.new_constraint();
      for (CUDFVersionedPackageListIterator ipkg = (*iup).remove_set.begin(); ipkg != (*iup).remove_set.end(); ipkg++) {
	if (solver.get_constraint_coeff(*ipkg) == 0) solver.set_constraint_coeff(*ipkg, +1);
	//printf("upgrade => remove %s\n", (*ipkg)->name);
      }
      solver.add_constraint_eq(0);

      // Force the installation of the upgraded packages
      int var_rank = (*iup).first_var_rank;
      bool has_mvp = false;
      solver.new_constraint();
      for (CUDFVersionedProviderListIterator ivpl = (*iup).upgrade_set.begin(); ivpl != (*iup).upgrade_set.end(); ivpl++)
	if (ivpl->second.size() == 1)
	  solver.set_constraint_coeff(ivpl->second.front(), +1);
	else {
	  has_mvp = true;
	  solver.set_constraint_coeff(var_rank++, +1);
	}
      solver.add_constraint_eq(1);
      if (has_mvp) {
	int var_rank = (*iup).first_var_rank;
	int size;
	for (CUDFVersionedProviderListIterator ivpl = (*iup).upgrade_set.begin(); ivpl != (*iup).upgrade_set.end(); ivpl++)
	  if ((size = ivpl->second.size()) > 1) {
	    solver.new_constraint();
	    for (CUDFProviderListIterator kpkg = ivpl->second.begin(); kpkg != ivpl->second.end(); kpkg++)
	      solver.set_constraint_coeff(*kpkg, +1);
	    solver.set_constraint_coeff(var_rank, -1);
	    solver.add_constraint_geq(0);
	    solver.new_constraint();
	    for (CUDFProviderListIterator kpkg = ivpl->second.begin(); kpkg != ivpl->second.end(); kpkg++)
	      solver.set_constraint_coeff(*kpkg, -1);
	    solver.set_constraint_coeff(var_rank, size);
	    solver.add_constraint_geq(0);
	    var_rank++;
	}
      }
    }
  }

  // Handling packages constraints
  if (all_packages.size() > 0)
    for (CUDFVersionedPackageListIterator ipkg = all_packages.begin(); ipkg != all_packages.end(); ipkg++) {
      // Depends
      if ((*ipkg)->depends != (CUDFVpkgFormula *)NULL) {
	bool has_coeff = false;
	// Conjunction of dependencies
	for (CUDFVpkgFormulaIterator anddeps = (*ipkg)->depends->begin(); anddeps != (*ipkg)->depends->end(); anddeps++) {
	  bool self_depend = false;
	  solver.new_constraint();
	  // Disjunction of dependencies
	  for (CUDFVpkgListIterator ordeps = (*anddeps)->begin(); ordeps != (*anddeps)->end(); ordeps++) {
	    CUDFVirtualPackage *vpackage = (*ordeps)->virtual_package;
	    a_compptr comp = get_comparator((*ordeps)->op);
	    // It depends from all the right versions of the package
	    if (vpackage->all_versions.size() > 0) {
	      for (CUDFVersionedPackageSetIterator jpkg = vpackage->all_versions.begin(); jpkg != vpackage->all_versions.end(); jpkg++)
		if (comp((*jpkg)->version, (*ordeps)->version)) {
		  if ((*jpkg) == (*ipkg)) { // Then, the dependency is always checked
		    self_depend = true;
		    has_coeff = false;
		    break;
		  } else if (solver.get_constraint_coeff(*jpkg) == 0) {
		    has_coeff = true;
		    solver.set_constraint_coeff(*jpkg, +1);
		  }
		}
	    }
	    // as well as from all the providers 
	    if ((! self_depend) && (vpackage->providers.size() > 0)) { 
	      for (CUDFProviderListIterator jpkg = vpackage->providers.begin(); jpkg != vpackage->providers.end(); jpkg++)
		if ((*jpkg) == (*ipkg)) { // Then, the dependency is always checked
		  self_depend = true;
		  has_coeff = false;
		  break;
		} else if (solver.get_constraint_coeff(*jpkg) == 0) {
		  has_coeff = true;
		  solver.set_constraint_coeff(*jpkg, +1);
		}
	    }
	    // as well as from all the versioned providers with the right version
	    if (! self_depend)
	      for (CUDFVersionedProviderListIterator jpkg = vpackage->versioned_providers.begin(); 
		   jpkg != vpackage->versioned_providers.end(); jpkg++)
		if (self_depend)
		  break;
		else if (comp(jpkg->first, (*ordeps)->version))
		  for (CUDFProviderListIterator kpkg = jpkg->second.begin(); kpkg != jpkg->second.end(); kpkg++) 
		    if ((*kpkg) == (*ipkg)) { // Then, the dependency is always checked
		      self_depend = true;
		      has_coeff = false;
		      break;
		    } else if (solver.get_constraint_coeff(*kpkg) == 0) {
		      has_coeff = true;
		      solver.set_constraint_coeff(*kpkg, +1);
		    }
	  }
	  if (has_coeff) {
	    solver.set_constraint_coeff(*ipkg, -1);
	    solver.add_constraint_geq(0);
	  } else if (!self_depend) { 
	    solver.set_constraint_coeff(*ipkg, 1);
	    solver.add_constraint_eq(0);
	  }
	}
      }

      // Conflicts
      if ((*ipkg)->conflicts != (CUDFVpkgList *)NULL) {
	if (generate_no_lp_conflict) {
	  for (CUDFVpkgListIterator ipkgop = (*ipkg)->conflicts->begin();  ipkgop != (*ipkg)->conflicts->end(); ipkgop++) {
	    CUDFVirtualPackage *vpackage = (*ipkgop)->virtual_package;
	    a_compptr comp = get_comparator((*ipkgop)->op);
	    if (vpackage->all_versions.size() > 0) {
	      // It conflicts with all the right versions of the package
	      for (CUDFVersionedPackageSetIterator jpkg = vpackage->all_versions.begin(); jpkg != vpackage->all_versions.end(); jpkg++) 
		if (((*jpkg) != (*ipkg)) && (comp((*jpkg)->version, (*ipkgop)->version))) {
		  solver.new_constraint();
		  solver.set_constraint_coeff(*ipkg, -1);
		  solver.set_constraint_coeff(*jpkg, -1);
		  solver.add_constraint_geq(-1);
		}
	    }
	    // as well as with all the providers
	    if (vpackage->providers.size() > 0) {
	      for (CUDFProviderListIterator jpkg = vpackage->providers.begin(); jpkg != vpackage->providers.end(); jpkg++)
		if (((*jpkg) != (*ipkg))) {
		  solver.new_constraint();
		  solver.set_constraint_coeff(*ipkg, -1);
		  solver.set_constraint_coeff(*jpkg, -1);
		  solver.add_constraint_geq(-1);
		}
	    }
	    // as well as with all the versioned providers with the right version
	    for (CUDFVersionedProviderListIterator jpkg = vpackage->versioned_providers.begin(); 
		 jpkg != vpackage->versioned_providers.end(); jpkg++)
	      if (comp(jpkg->first, (*ipkgop)->version))
		for (CUDFProviderListIterator kpkg = jpkg->second.begin(); kpkg != jpkg->second.end(); kpkg++) 
		  if (((*kpkg) != (*ipkg))) {
		    solver.new_constraint();
		    solver.set_constraint_coeff(*ipkg, -1);
		    solver.set_constraint_coeff(*kpkg, -1);
		    solver.add_constraint_geq(-1);
		  }
	  }
	} else {
	  int nb_coeff = 0;
	  
	  solver.new_constraint();
	
	  for (CUDFVpkgListIterator ipkgop = (*ipkg)->conflicts->begin();  ipkgop != (*ipkg)->conflicts->end(); ipkgop++) {
	    CUDFVirtualPackage *vpackage = (*ipkgop)->virtual_package;
	    a_compptr comp = get_comparator((*ipkgop)->op);
	    if (vpackage->all_versions.size() > 0) {
	      // It conflicts with all the right versions of the package
	      for (CUDFVersionedPackageSetIterator jpkg = vpackage->all_versions.begin(); jpkg != vpackage->all_versions.end(); jpkg++) 
		if (((*jpkg) != (*ipkg)) && (comp((*jpkg)->version, (*ipkgop)->version))&& (solver.get_constraint_coeff(*jpkg) == 0)) {
		  solver.set_constraint_coeff(*jpkg, -1);
		  nb_coeff++;
		}
	    }
	    // as well as with all the providers
	    if (vpackage->providers.size() > 0) {
	      for (CUDFProviderListIterator jpkg = vpackage->providers.begin(); jpkg != vpackage->providers.end(); jpkg++)
		if (((*jpkg) != (*ipkg)) && (solver.get_constraint_coeff(*jpkg) == 0)) {
		  solver.set_constraint_coeff(*jpkg, -1);
		  nb_coeff++;
		}
	    }
	    // as well as with all the versioned providers with the right version
	    for (CUDFVersionedProviderListIterator jpkg = vpackage->versioned_providers.begin(); 
		 jpkg != vpackage->versioned_providers.end(); jpkg++)
	      if (comp(jpkg->first, (*ipkgop)->version))
		for (CUDFProviderListIterator kpkg = jpkg->second.begin(); kpkg != jpkg->second.end(); kpkg++) 
		  if (((*kpkg) != (*ipkg)) && (solver.get_constraint_coeff(*kpkg) == 0)) {
		    solver.set_constraint_coeff(*kpkg, -1);
		    nb_coeff++;
		  }
	  }
	  if (nb_coeff > 0) {
	    solver.set_constraint_coeff(*ipkg, -nb_coeff);
	    solver.add_constraint_geq(-nb_coeff);
	  }
	}
      }

      // Keep (if and only if installed)
      if ((*ipkg)->installed)
	switch((*ipkg)->keep) {
	case keep_none: break;
	case keep_feature: // Preserve all the provided features
	  if ((*ipkg)->provides != (CUDFVpkgList *)NULL) {
	    for (CUDFVpkgListIterator ipkgop = (*ipkg)->provides->begin(); ipkgop != (*ipkg)->provides->end(); ipkgop++) {
	      CUDFVirtualPackage *vpackage = (*ipkgop)->virtual_package;
	      a_compptr comp = get_comparator((*ipkgop)->op);
	      solver.new_constraint();
	      if (vpackage->all_versions.size() > 0)
		for (CUDFVersionedPackageSetIterator jpkg = vpackage->all_versions.begin(); jpkg != vpackage->all_versions.end(); jpkg++)
		  if (comp((*jpkg)->version, (*ipkgop)->version)) solver.set_constraint_coeff(*jpkg, +1);
	      if (vpackage->providers.size() > 0)
		for (CUDFProviderListIterator jpkg = vpackage->providers.begin(); jpkg != vpackage->providers.end(); jpkg++)
		  if (solver.get_constraint_coeff(*jpkg) == 0) solver.set_constraint_coeff(*jpkg, +1);
	      for (CUDFVersionedProviderListIterator jpkg = vpackage->versioned_providers.begin(); 
		   jpkg != vpackage->versioned_providers.end(); jpkg++)
		if (comp(jpkg->first, (*ipkgop)->version))
		  for (CUDFProviderListIterator kpkg = jpkg->second.begin(); kpkg != jpkg->second.end(); kpkg++) 
		    if (solver.get_constraint_coeff(*kpkg) == 0) solver.set_constraint_coeff(*kpkg, +1);
	      solver.add_constraint_geq(+1);
	    }
	  }
	  break;
	case keep_package: // Preserve at least one version of the package
	  if (((*ipkg)->virtual_package->all_versions.size() > 0) && (! keep_package_handled[(*ipkg)->virtual_package->rank+1])) {
	    CUDFVirtualPackage *vpackage = (*ipkg)->virtual_package;
	    solver.new_constraint();
	    if (vpackage->all_versions.size() > 0)  // Should not make sense
	      for (CUDFVersionedPackageSetIterator jpkg = vpackage->all_versions.begin(); jpkg != vpackage->all_versions.end(); jpkg++)
		  solver.set_constraint_coeff(*jpkg, +1);
	    solver.add_constraint_geq(+1);
	    keep_package_handled[(*ipkg)->virtual_package->rank+1] = true;
	  }
	  break;
	case keep_version: // Preserve the current version
	  solver.new_constraint();
	  solver.set_constraint_coeff(*ipkg, +1);
	  solver.add_constraint_eq(+1);
	  break;
	}
    }

  solver.end_add_constraints();

  return 0;
}

