
/*******************************************************/
/* (c) Claude Michel I3S (UNSA-CNRS) 2009,2010         */
/* Distributed under GPLv3                             */
/*******************************************************/


#ifndef _CPLEX_SOLVER_H
#define _CPLEX_SOLVER_H

#include <abstract_solver.h>
#include <ilcplex/cplex.h>

class cplex_solver: public abstract_solver {
 public:
  int init_solver(CUDFVersionedPackageList *all_versioned_packages, int other_vars);
  int writelp(char *filename);
  int solve();
  double objective_value();
  int init_solutions();
  double get_solution(CUDFVersionedPackage *package);

  int begin_objective(void);
  double get_obj_coeff(CUDFVersionedPackage *package);
  int set_obj_coeff(CUDFVersionedPackage *package, double value);
  int set_obj_coeff(int rank, double value);
  int end_objective(void);
  int new_objective(void);
  int add_objective(void);

  int begin_add_constraints(void);
  int new_constraint(void);
  double get_constraint_coeff(CUDFVersionedPackage *package);
  int set_constraint_coeff(CUDFVersionedPackage *package, double value);
  int set_constraint_coeff(int rank, double value);
  int add_constraint_geq(double bound);
  int add_constraint_leq(double bound);
  int add_constraint_eq(double bound);
  int end_add_constraints(void);

  CPXENVptr env; 
  CPXLPptr lp;
  CUDFVersionedPackageList *all_versioned_packages;

  double *coefficient;
  int *index;
  int rank;
  int nb_packages;
  int nb_vars;
  int status;
  double *lb;
  double *ub;
  char *vartype;
  char **varname;
  int begin[2];
  double rhs[1];
  char sense[1];

  int nb_objectives;
  double *objs[20];
  int *obj_index;

  double *solution;

  cplex_solver(void) {
    all_versioned_packages = (CUDFVersionedPackageList *)NULL;
    rank = 0;
    coefficient = (double *)NULL;
    index = (int *)NULL;
    solution = (double *)NULL;
    nb_objectives = 0;
    obj_index = (int *)NULL;
  }
};

#endif
