
/*******************************************************/
/* (c) Claude Michel I3S (UNSA-CNRS) 2009,2010         */
/* Distributed under GPLv3                             */
/*******************************************************/


#include <cpo_solver.h>
#include <ilcp/cp.h>
#include <ilconcert/iloxmlcontext.h>

abstract_solver *new_cpo_solver() { return new cpo_solver(); }

int cpo_solver::init_solver(CUDFVersionedPackageList *all_versioned_packages, int other_vars) {

  nb_packages = all_versioned_packages->size();
  nb_vars = nb_packages + other_vars;
  this->all_versioned_packages = all_versioned_packages;

  /* Initialize the CP environment */
  env = new IloEnv();

  if ( env == (IloEnv *)NULL ) {
    fprintf (stderr, "Could not open CPO environment.\n");
    exit(-1);
  }  
  
  /* Create the model */

  try {
    vars = new IloBoolVarArray(*env, nb_vars);
    model = new IloModel(*env);
  }  catch (IloException& ex) {
    env->out() << "CPO Error: " << ex << std::endl;
    exit(-1);
  }


  coefficient = (double *)malloc(nb_vars*sizeof(double));
  index = (int *)malloc(nb_vars*sizeof(int));

  if ((coefficient == (double *)NULL) ||
      (index  == (int *)NULL))
    return -1;
  else
    return 0;
}

int cpo_solver::writelp(char *filename) { return 0; }

int cpo_solver::solve() {
  try {
    //    IloXmlContext context(*env);
    //    context.writeModel(*model, "cpomodel.xml");
    //    context.end();
    //   env->out() << *model << std::endl;
    cp = new IloCP(*model);
    if (cp->solve()) return 1;
  }  catch (IloException& ex) {
    env->out() << "CPO Error: " << ex << std::endl;
  }
  return 0;
}

double cpo_solver::objective_value() { 
  return cp->getValue(*obj);
  //  return 0;
}

int cpo_solver::init_solutions() {
  return 0; 
}

double cpo_solver::get_solution(CUDFVersionedPackage *package) { return cp->getValue((*vars)[package->rank]); }

int cpo_solver::begin_objective(void) { 
  int i = 0;
  for (CUDFVersionedPackageListIterator ipkg = all_versioned_packages->begin(); ipkg != all_versioned_packages->end(); ipkg++) {
    coefficient[i] = 0;
    i++;
  }
  for (i = nb_packages; i < nb_vars; i++) {
    coefficient[i] = 0;
  }

  return 0; 
}

double cpo_solver::get_obj_coeff(CUDFVersionedPackage *package) { return coefficient[package->rank]; }

int cpo_solver::set_obj_coeff(CUDFVersionedPackage *package, double value) { coefficient[package->rank] = value; return 0; }

int cpo_solver::end_objective(void) {
  // Build the objective function
  int retval = 0;
  try {
    obj = new IloExpr(*env);
    for(int i = 0; i < nb_vars; i++)
      if (coefficient[i] != 0) 
	*obj += coefficient[i] * (*vars)[i];
    model->add(IloMinimize(*env, *obj, "obj"));
  }  catch (IloException& ex) {
    env->out() << "CPO Error: " << ex << std::endl;
    retval = 1;
  }

  return retval;
}

int cpo_solver::begin_add_constraints(void) { return 0; }

int cpo_solver::new_constraint(void) { rank = 0; return 0; }

double cpo_solver::get_constraint_coeff(CUDFVersionedPackage *package) { 
  for(int i =0; i < rank; i++)
    if (index[i] == package->rank) return coefficient[i];
  return 0;
}

int cpo_solver::set_constraint_coeff(CUDFVersionedPackage *package, double value) { 
  coefficient[rank] = value;
  index[rank] = package->rank;
  rank++;
  return 0;
}

int cpo_solver::set_constraint_coeff(int therank, double value) { 
  coefficient[rank] = value;
  index[rank] = therank;
  rank++;
  return 0;
}

int cpo_solver::add_constraint_geq(double bound) {
  int retval = 0;
  if (rank > 0) {
    try {
      IloExpr expr(*env);
      for(int i =0; i < rank; i++)
	expr += coefficient[i] * (*vars)[index[i]];
      model->add(expr >= bound);
    }  catch (IloException& ex) {
      env->out() << "CPO Error: " << ex << std::endl;
      retval = 1;
    }
  }

  return retval;
}

int cpo_solver::add_constraint_leq(double bound) {
  int retval = 0;
  if (rank > 0) {
    try {
      IloExpr expr(*env);
      for(int i =0; i < rank; i++)
	expr += coefficient[i] * (*vars)[index[i]];
      model->add(expr <= bound);
    }  catch (IloException& ex) {
      env->out() << "CPO Error: " << ex << std::endl;
      retval = 1;
    }
  }

  return retval;
}

int cpo_solver::add_constraint_eq(double bound) {
  int retval = 0;
  if (rank > 0) {
    try {
      IloExpr expr(*env);
      for(int i =0; i < rank; i++)
	expr += coefficient[i] * (*vars)[index[i]];
      model->add(expr == bound);
    }  catch (IloException& ex) {
      env->out() << "CPO Error: " << ex << std::endl;
      retval = 1;
    }
  }

  return retval;
}

int cpo_solver::end_add_constraints(void) { return 0; }
