
/*******************************************************/
/* (c) Claude Michel I3S (UNSA-CNRS) 2009,2010         */
/* Distributed under GPLv3                             */
/*******************************************************/


#ifndef _CPO_SOLVER_H
#define _CPO_SOLVER_H

#include <abstract_solver.h>
#include <ilcp/cp.h>

class cpo_solver: public abstract_solver {
 public:
  int init_solver(CUDFVersionedPackageList *all_versioned_packages, int other_vars);
  int writelp(char *filename);
  int solve();
  double objective_value();
  int init_solutions();
  double get_solution(CUDFVersionedPackage *package);

  int begin_objective(void);
  double get_obj_coeff(CUDFVersionedPackage *package);
  int set_obj_coeff(CUDFVersionedPackage *package, double value);
  int end_objective(void);

  int begin_add_constraints(void);
  int new_constraint(void);
  double get_constraint_coeff(CUDFVersionedPackage *package);
  int set_constraint_coeff(CUDFVersionedPackage *package, double value);
  int set_constraint_coeff(int rank, double value);
  int add_constraint_geq(double bound);
  int add_constraint_leq(double bound);
  int add_constraint_eq(double bound);
  int end_add_constraints(void);

  CUDFVersionedPackageList *all_versioned_packages;

  IloEnv *env;
  IloModel *model;
  IloCP *cp;
  IloBoolVarArray *vars;
  IloExpr *obj;

  double *coefficient;
  int *index;
  int rank;
  int nb_packages;
  int nb_vars;
  int status;

  double *solution;

  cpo_solver(void) {
    all_versioned_packages = (CUDFVersionedPackageList *)NULL;
    env = (IloEnv *)NULL;
    model = (IloModel *)NULL;
    cp = (IloCP *)NULL;
    vars = (IloBoolVarArray *)NULL;
    rank = 0;
    coefficient = (double *)NULL;
    index = (int *)NULL;
    solution = (double *)NULL;
  }
};

#endif
