
/*******************************************************/
/* (c) Claude Michel I3S (UNSA-CNRS) 2009,2010         */
/* Distributed under GPLv3                             */
/*******************************************************/


#include <glpk_solver.h>


abstract_solver *new_glpk_solver(bool use_exact) { return new glpk_solver(use_exact); }

int glpk_solver::init_solver(CUDFVersionedPackageList *all_versioned_packages, int other_vars) {
  nb_packages = all_versioned_packages->size();
  nb_vars = nb_packages+other_vars;

  this->all_versioned_packages = all_versioned_packages;

  lp = glp_create_prob();
  glp_add_cols(lp, nb_vars);

  coefficient = (double *)malloc((nb_vars+1)*sizeof(double));
  index = (int *)malloc((nb_vars+1)*sizeof(int));

  if ((coefficient == (double *)NULL) || (index  == (int *)NULL))
    return -1;
  else
    return 0;
}

int glpk_solver::writelp(char *filename) { glp_write_lp(lp, NULL, filename); return 0; }

int glpk_solver::solve() {
  int status = 0;
  //  glp_smcp simplex_params;
  glp_iocp mip_params;

  //  glp_init_smcp(&simplex_params);
  //  simplex_params.presolve = GLP_ON;

  glp_init_iocp(&mip_params);
  mip_params.gmi_cuts = GLP_ON;
  mip_params.mir_cuts = GLP_ON;
  mip_params.cov_cuts = GLP_ON;
  mip_params.clq_cuts = GLP_ON;
  mip_params.presolve = GLP_ON;
  mip_params.binarize = GLP_ON;

  //  glp_scale_prob(lp, GLP_SF_AUTO);
  //  glp_adv_basis(lp, 0);
  glp_cpx_basis(lp);
  //  if (use_exact) status = glp_exact(lp, &simplex_params); else  status = glp_simplex(lp, &simplex_params);
  //  if (use_exact) status = glp_exact(lp, NULL); else  status = glp_simplex(lp, NULL);
  if (status == 0) status = glp_intopt(lp, &mip_params);
  if (status == 0)  return 1; else return 0;
}

double glpk_solver::objective_value() { return glp_mip_obj_val(lp); }

int glpk_solver::init_solutions() { return 0; }

double glpk_solver::get_solution(CUDFVersionedPackage *package) { return glp_mip_col_val(lp, package->rank+1); }

int glpk_solver::begin_objective(void) { 
  glp_set_obj_dir(lp, GLP_MIN);  // Problem is minimization
  return 0; 
}

double glpk_solver::get_obj_coeff(CUDFVersionedPackage *package) { return glp_get_obj_coef(lp, package->rank+1); }

int glpk_solver::set_obj_coeff(CUDFVersionedPackage *package, double value) { glp_set_obj_coef(lp, package->rank+1, value); return 0; }
int glpk_solver::set_obj_coeff(int rank, double value) { glp_set_obj_coef(lp, rank+1, value); return 0; }

int glpk_solver::end_objective(void) {
  int i = 1;
  for (CUDFVersionedPackageListIterator ipkg = all_versioned_packages->begin(); ipkg != all_versioned_packages->end(); ipkg++) {
    glp_set_col_bnds(lp, i, GLP_DB, 0, 1);  // Set bounds to [0, 1]
    glp_set_col_name(lp, i, (*ipkg)->versioned_name); // Set the colunm name
    glp_set_col_kind(lp, i, GLP_BV); // It is a binary variable ...
    i++;
  }
  for (i = nb_packages+1; i <= nb_vars; i++) {
    char *name;
    char buffer[20];

    sprintf(buffer, "x%d", i);
    if ((name = (char *)malloc(strlen(buffer)+1)) == (char *)NULL) {
      fprintf(stderr, "CUDF error: can not alloc memory for variable name in glpk_solver::end_objective.\n");
      exit(-1);
    }
    strcpy(name, buffer);

    glp_set_col_bnds(lp, i, GLP_DB, 0, 1);  // Set bounds to [0, 1]
    glp_set_col_name(lp, i, name); // Set the colunm name
    glp_set_col_kind(lp, i, GLP_BV); // It is a binary variable ...
  }
  return 0;
}

int glpk_solver::begin_add_constraints(void) { return 0; }

int glpk_solver::new_constraint(void) { rank = 1; return 0; }

double glpk_solver::get_constraint_coeff(CUDFVersionedPackage *package) { 
  int prank = package->rank + 1;
  for (int i = 1; i < rank; i++)
    if (index[i] == prank) return coefficient[i];
  return 0;
}

int glpk_solver::set_constraint_coeff(CUDFVersionedPackage *package, double value) { 
  coefficient[rank] = value;
  index[rank] = package->rank + 1;
  rank++;
  return 0;
}

int glpk_solver::set_constraint_coeff(int therank, double value) { 
  coefficient[rank] = value;
  index[rank] = therank + 1;
  rank++;
  return 0;
}

int glpk_solver::add_constraint_geq(double bound) {
  if (rank > 1 ) {
    int irow = glp_add_rows(lp, 1);
    glp_set_row_bnds(lp, irow, GLP_LO, bound, 0);
    glp_set_mat_row(lp, irow, rank-1, index, coefficient);
  }
  return 0;
}

int glpk_solver::add_constraint_leq(double bound) {
  if (rank > 1 ) {
    int irow = glp_add_rows(lp, 1);
    glp_set_row_bnds(lp, irow, GLP_UP, 0, bound);
    glp_set_mat_row(lp, irow, rank-1, index, coefficient);
  }
  return 0;
}

int glpk_solver::add_constraint_eq(double bound) {
  if (rank > 1 ) {
    int irow = glp_add_rows(lp, 1);
    glp_set_row_bnds(lp, irow, GLP_FX, bound, bound);
    glp_set_mat_row(lp, irow, rank-1, index, coefficient);
  }
  return 0;
}

int glpk_solver::end_add_constraints(void) { glp_write_lp(lp, NULL, "glpkpbs.lp"); return 0; }
