
/*******************************************************/
/* (c) Claude Michel I3S (UNSA-CNRS) 2009,2010         */
/* Distributed under GPLv3                             */
/*******************************************************/


#ifndef _GLPK_SOLVER_H
#define _GLPK_SOLVER_H

#include <abstract_solver.h>
#include <glpk.h>

class glpk_solver: public abstract_solver {
 public:
  int init_solver(CUDFVersionedPackageList *all_versioned_packages, int other_vars);
  int writelp(char *filename);
  int solve();
  double objective_value();
  int init_solutions();
  double get_solution(CUDFVersionedPackage *package);

  int begin_objective(void);
  double get_obj_coeff(CUDFVersionedPackage *package);
  int set_obj_coeff(CUDFVersionedPackage *package, double value);
  int set_obj_coeff(int rank, double value);
  int end_objective(void);

  int begin_add_constraints(void);
  int new_constraint(void);
  double get_constraint_coeff(CUDFVersionedPackage *package);
  int set_constraint_coeff(CUDFVersionedPackage *package, double value);
  int set_constraint_coeff(int rank, double value);
  int add_constraint_geq(double bound);
  int add_constraint_leq(double bound);
  int add_constraint_eq(double bound);
  int end_add_constraints(void);

  glp_prob *lp;
  CUDFVersionedPackageList *all_versioned_packages;
  int nb_packages;
  int nb_vars;

  double *coefficient;
  int *index;
  int rank;

  bool use_exact;

  glpk_solver(bool use_exact) {
    this->use_exact = use_exact;
    lp = (glp_prob *)NULL;
    all_versioned_packages = (CUDFVersionedPackageList *)NULL;
    rank = 1;
    coefficient = (double *)NULL;
    index = (int *)NULL;
  }
};

#endif
