
/*******************************************************/
/* (c) Claude Michel I3S (UNSA-CNRS) 2009,2010         */
/* Distributed under GPLv3                             */
/*******************************************************/


#include <lpsolve_solver.h>


abstract_solver *new_lpsolve_solver() { return new lpsolve_solver(); }

int lpsolve_solver::init_solver(CUDFVersionedPackageList *all_versioned_packages, int other_vars) {
  nb_packages = all_versioned_packages->size();
  nb_vars = nb_packages + other_vars;

  this->all_versioned_packages = all_versioned_packages;

  if ((lp = make_lp(0, nb_vars)) == NULL) {
    fprintf(stderr, "Cannot create the lpsolve solver.\n");
    exit(-1);
  }

  coefficient = (double *)malloc((nb_vars+1)*sizeof(double));
  index = (int *)malloc((nb_vars+1)*sizeof(int));

  if ((coefficient == (double *)NULL) || (index  == (int *)NULL))
    return -1;
  else
    return 0;
}

int lpsolve_solver::writelp(char *filename) { write_lp(lp, filename); return 0; }

int lpsolve(lprec *lp) { return solve(lp); }

int lpsolve_solver::solve() {
  int status;
  set_presolve(lp, PRESOLVE_ROWS | PRESOLVE_COLS | PRESOLVE_LINDEP | 
	       PRESOLVE_REDUCEMIP | PRESOLVE_REDUCEGCD | PRESOLVE_PROBEFIX | PRESOLVE_PROBEREDUCE, get_presolveloops(lp));
  status = lpsolve(lp);
  if ((status == OPTIMAL) || (status == SUBOPTIMAL) || (status == PRESOLVED)) return 1; else return 0;
}

double lpsolve_solver::objective_value() { return get_objective(lp); }

int lpsolve_solver::init_solutions() { get_ptr_variables(lp, &solution); return 0; }

double lpsolve_solver::get_solution(CUDFVersionedPackage *package) { return solution[package->rank]; }

int lpsolve_solver::begin_objective(void) { 
  set_minim(lp);  // Problem is minimization
  return 0; 
}

double lpsolve_solver::get_obj_coeff(CUDFVersionedPackage *package) { return coefficient[package->rank+1]; }

int lpsolve_solver::set_obj_coeff(CUDFVersionedPackage *package, double value) { coefficient[package->rank+1] = value; return 0; }
int lpsolve_solver::set_obj_coeff(int rank, double value) { coefficient[rank+1] = value; return 0; }

int lpsolve_solver::end_objective(void) {

  // Build the objective function
  if (! set_obj_fn(lp, coefficient)) {
    fprintf(stderr, "generate_milp: cannot create objective function.\n");
    exit(-1);
  }

  // Use names for colunms
  set_use_names(lp, FALSE, TRUE);
  for (CUDFVersionedPackageListIterator ipkg = all_versioned_packages->begin(); ipkg != all_versioned_packages->end(); ipkg++) {
    int i = (*ipkg)->rank+1;
    if ((! set_bounds(lp, i, 0, 1)) ||
	(! set_binary(lp, i, TRUE)) ||
	(! set_col_name(lp, i, (*ipkg)->versioned_name))) {
      fprintf(stderr, "generate_milp: cannot set type, bounds or name for objective function.\n");
      exit(-1);
    }
  }
  for (int i = nb_packages+1; i <= nb_vars; i++) {
    char *name;
    char buffer[20];

    sprintf(buffer, "x%d", i);
    if ((name = (char *)malloc(strlen(buffer)+1)) == (char *)NULL) {
      fprintf(stderr, "CUDF error: can not alloc memory for variable name in glpk_solver::end_objective.\n");
      exit(-1);
    }
    strcpy(name, buffer);

    if ((! set_bounds(lp, i, 0, 1)) ||
	(! set_binary(lp, i, TRUE)) ||
	(! set_col_name(lp, i, name))) {
      fprintf(stderr, "generate_milp: cannot set type, bounds or name for objective function.\n");
      exit(-1);
    }
  }

  return 0;
}

int lpsolve_solver::begin_add_constraints(void) { 
  set_add_rowmode(lp, TRUE);
  return 0; 
}

int lpsolve_solver::new_constraint(void) { rank = 0; return 0; }

double lpsolve_solver::get_constraint_coeff(CUDFVersionedPackage *package) {
  for (int i = 0; i < rank; i++)
    if (index[i] == package->rank + 1) return coefficient[i];
  return 0;
}

int lpsolve_solver::set_constraint_coeff(CUDFVersionedPackage *package, double value) { 
  coefficient[rank] = value;
  index[rank] = package->rank + 1;
  rank++;
  return 0;
}

int lpsolve_solver::set_constraint_coeff(int therank, double value) { 
  coefficient[rank] = value;
  index[rank] = therank + 1;
  rank++;
  return 0;
}

int lpsolve_solver::add_constraint_geq(double bound) {
  if (rank > 0) add_constraintex(lp, rank, coefficient, index, GE, bound);
  return 0;
}

int lpsolve_solver::add_constraint_leq(double bound) {
  if (rank > 0) add_constraintex(lp, rank, coefficient, index, LE, bound);
  return 0;
}

int lpsolve_solver::add_constraint_eq(double bound) {
  if (rank > 0) add_constraintex(lp, rank, coefficient, index, EQ, bound);
  return 0;
}

int lpsolve_solver::end_add_constraints(void) { 
  set_add_rowmode(lp, FALSE);
  write_lp(lp, (char *)"lpsolvepbs.lp"); 
  return 0; 
}
