
/*******************************************************/
/* (c) Claude Michel I3S (UNSA-CNRS) 2009,2010         */
/* Distributed under GPLv3                             */
/*******************************************************/


#include <pblib_solver.h>
#include <stdio.h>
#include <stdlib.h>

abstract_solver *new_pblib_solver(char *pb_solver) { return new pblib_solver(pb_solver); }

int pblib_solver::init_solver(CUDFVersionedPackageList *all_versioned_packages, int other_vars) {

  nb_packages = all_versioned_packages->size();
  nb_vars = nb_packages+other_vars;

  this->all_versioned_packages = all_versioned_packages;

  nb_constraints = 0;
  usenames = false;
  mult = ' ';

  for (int i = 0; i < ((int)strlen(pb_solver) - 7); i++)
    if ((pb_solver[i] == 'm') && (strncmp(pb_solver+i, "minisat", 7) == 0)) { mult = '*'; break; }

  coefficient = (double *)malloc(nb_vars*sizeof(double));
  index = (int *)malloc(nb_vars*sizeof(int));
  vname = (char **)malloc(nb_vars*sizeof(char *));
  solution = (double *)malloc(nb_vars*sizeof(double));

  pbfile = fopen("pblib.opb", "w");

  if ((coefficient == (double *)NULL) || 
      (index  == (int *)NULL) || 
      (vname == (char **)NULL) || 
      (solution == (double *)NULL) ||
      (pbfile == (FILE *)NULL))
    return -1;
  else
    return 0;
}

int pblib_solver::writelp(char *filename) { return 0; }

int pblib_solver::solve() {
  int status = 0;
  int rank, iobjval;
  char command[1024];
  FILE *fsol = (FILE *)NULL;
  bool is_scip = false;

  for (int i = 0; i < ((int)strlen(pb_solver) - 4); i++)
    if ((pb_solver[i] == 's') && (strncmp(pb_solver+i, "scip", 4) == 0)) { is_scip = true; break; }

  if (is_scip) {
    int state = 1;
    size_t size = 2048;
    char buff[2048]; char *buffer = buff;

    sprintf(command, "%s -f pblib.opb | tee pbsolver.out", pb_solver);
    system(command);
    if ((fsol = fopen("pbsolver.out", "r")) == (FILE *)NULL) {
      fprintf(stderr, "Cannot open solution file \"pbsolver.ou\".\n");
      exit(-1);
    }

    for (int i = 0; i < nb_packages; i++) solution[i] = 0;

    while (! feof(fsol)) { /* seek : primal solution: */
      int nbread = getline((char **)&buffer, &size, fsol);
      if ((state == 1) && (buffer[0] == 'p') && (nbread >= 16) && (strncmp(buffer, "primal solution:", 16) == 0)) {
	nbread = getline((char **)&buffer, &size, fsol); /* forget next line */
	nbread = getline((char **)&buffer, &size, fsol); /* forget next line */
	state = 2;
      } else if (state == 2) {
	if ((buffer[0] == 'n') && (nbread >= 21) && (strncmp(buffer, "no solution available", 21) == 0))
	  break;
	else if ((buffer[0] == 'o') && (nbread >= 16) && (strncmp(buffer, "objective value:", 16) == 0)) {
	  int i = 17;
	  while ((buffer[i] != '-') && ((buffer[i] < '0') || (buffer[i] > '9'))) i++;
	  if (sscanf(buffer+i, "%d", &iobjval) > 0) objval = iobjval;
	  status = 1;
	  state = 3;
	}
      } else if (state == 3) {
	if (buffer[0] == ' ') 
	  break;
	else if ((buffer[0] == 'x') && (sscanf(buffer+1, "%d", &rank) > 0)) 
	  solution[rank-1] = 1;
      }
    }

  } else {
    sprintf(command, "%s pblib.opb | tee pbsolver.out", pb_solver);
    system(command);

    if ((fsol = fopen("pbsolver.out", "r")) == (FILE *)NULL) {
      fprintf(stderr, "Cannot open solution file \"pbsolver.ou\".\n");
      exit(-1);
    }

    for (int i = 0; i < nb_packages; i++) solution[i] = 0;

    while (! feof(fsol)) {
      char c = fgetc(fsol);
      switch (c) {
      case '\n':
	break;
      case 's': 
	fgetc(fsol);
	if (fgets(command, 1000, fsol) != NULL)
	  if (strncmp(command, "OPTIMUM FOUND", 12) == 0) status = 1;
	break;
      case 'v':
	while (! feof(fsol)) {
	  c = fgetc(fsol);
	  if (c == '\n')
	    break;
	  else if (c == '-')
	    while (! feof(fsol)) { if (fgetc(fsol) == ' ') break;}
	  else if (c == 'x')
	    if (fscanf(fsol, "%d", &rank) > 0) solution[rank-1] = 1;
	}
	break;
      case 'o':
	fgetc(fsol);
	if (fscanf(fsol, "%d", &iobjval) > 0) objval = iobjval;
      default:
	while ((! feof(fsol)) && (fgetc(fsol) != '\n'));
      }
    }
  }

  return status;
}

double pblib_solver::objective_value() { return objval; }

int pblib_solver::init_solutions() { return 0; }

double pblib_solver::get_solution(CUDFVersionedPackage *package) { return solution[package->rank]; }

int pblib_solver::begin_objective(void) { 
  for (int i = 0; i < nb_vars; i++) coefficient[i] = 0;
  return 0; 
}

double pblib_solver::get_obj_coeff(CUDFVersionedPackage *package) { return coefficient[package->rank]; }

int pblib_solver::set_obj_coeff(CUDFVersionedPackage *package, double value) { coefficient[package->rank] = value; return 0; }

int pblib_solver::set_obj_coeff(int rank, double value) { coefficient[rank] = value; return 0; }

int pblib_solver::end_objective(void) {
  fprintf(pbfile, "* #variable= %d #constraint=                         \n* a CUDF problem\nmin:", nb_vars);

  for (CUDFVersionedPackageListIterator ipkg = all_versioned_packages->begin(); ipkg != all_versioned_packages->end(); ipkg++) {
    int i = (*ipkg)->rank;
    if (coefficient[i] != 0) {
      if (usenames)
	fprintf(pbfile, " %+d%c%s", (int)coefficient[i], mult, (*ipkg)->versioned_name);
      else
	fprintf(pbfile, " %+d%cx%d", (int)coefficient[i], mult, i+1);
    }
  }
  for (int k = nb_packages; k < nb_vars; k++) 
    if (coefficient[k] != 0) fprintf(pbfile, " %+d%cx%d", (int)coefficient[k], mult, k+1);
  fprintf(pbfile, ";\n");

  return 0;
}

int pblib_solver::begin_add_constraints(void) { 
  return 0; 
}

int pblib_solver::new_constraint(void) { rank = 0; return 0; }

double pblib_solver::get_constraint_coeff(CUDFVersionedPackage *package) {
  for (int i = 0; i < rank; i++)
    if (index[i] == package->rank) return coefficient[i];
  return 0;
}

int pblib_solver::set_constraint_coeff(CUDFVersionedPackage *package, double value) { 
  coefficient[rank] = (int)value;
  index[rank] = package->rank;
  vname[rank] = package->versioned_name;
  rank++;
  return 0;
}

int pblib_solver::set_constraint_coeff(int therank, double value) { 
  coefficient[rank] = (int)value;
  index[rank] = therank;
  vname[rank] = (char *)NULL;
  rank++;
  return 0;
}

int pblib_solver::add_constraint_geq(double bound) {
  char *sep = (char *)"";

  if (rank > 0) {
    for (int i = 0; i < rank; i++) {
      if (usenames && (vname[i] != (char *)NULL))
	fprintf(pbfile, "%s%+d%c%s", sep, (int)coefficient[i], mult, vname[i]);
      else
	fprintf(pbfile, "%s%+d%cx%d", sep, (int)coefficient[i], mult, index[i]+1);
      sep = (char *)" ";
    }
    if (bound == 0) fprintf(pbfile, " >= 0;\n"); else fprintf(pbfile, " >= %+d;\n", (int)bound);
    nb_constraints++;
  }
  return 0;
}

int pblib_solver::add_constraint_leq(double bound) {
  char *sep = (char *)"";

  if (rank > 0) {
    for (int i = 0; i < rank; i++) {
      if (usenames && (vname[i] != (char *)NULL))
	fprintf(pbfile, "%s%+d%c%s", sep, (int)coefficient[i], mult, vname[i]);
      else
	fprintf(pbfile, "%s%+d%cx%d", sep, (int)coefficient[i], mult, index[i]+1);
      sep = (char *)" ";
    }
    if (bound == 0) fprintf(pbfile, " <= 0;\n"); else fprintf(pbfile, " <= %+d;\n", (int)bound);
    nb_constraints++;
  }
  return 0;
}

int pblib_solver::add_constraint_eq(double bound) {
  char *sep = (char *)"";

  if (rank > 0) {
    for (int i = 0; i < rank; i++) {
      if (usenames && (vname[i] != (char *)NULL))
	fprintf(pbfile, "%s%+d%c%s", sep, (int)coefficient[i], mult, vname[i]);
      else
	fprintf(pbfile, "%s%+d%cx%d", sep, (int)coefficient[i], mult, index[i]+1);
      sep = (char *)" ";
    }
    if (bound == 0) fprintf(pbfile, " = 0;\n"); else fprintf(pbfile, " = %+d;\n", (int)bound);
    nb_constraints++;
  }
  return 0;
}

int pblib_solver::end_add_constraints(void) { 
  rewind(pbfile);
  fprintf(pbfile, "* #variable= %d #constraint= %d", nb_vars, nb_constraints);
  fclose(pbfile);
  return 0; 
}
